"""Top-level package for centra-py-client."""

__author__ = """Yonatan Golan"""
__email__ = 'yonatan.golan@guardicore.com'
__version__ = '0.3.0'

import logging

from .centra_py_client import CentraClient as CentraClient  # noqa: F401
from .centra_py_client import CentraSession as CentraSession  # noqa: F401


logging.getLogger(__name__).addHandler(logging.NullHandler())



import logging
from typing import List

from centra_py_client.exceptions import ManagementAPIError
from centra_py_client.centra_session import CentraSession


class CentraClient:
    def __init__(self, centra_session: CentraSession):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.centra_session = centra_session

    def list_assets(self, **filt):
        return self.centra_session.json_query(self.centra_session.urljoin_api('assets'), params=filt)['objects']

    def add_label_to_assets(self, asset_ids: List[str], label_key: str, label_value: str) -> str:
        """
        Add a label to assets.
        Note - If the assets are already labeled with a label that has the provided label_key, they will be
        automatically removed from their current label and added to the new provided label.
        :param asset_ids: A list of asset ids to add the label to
        :param label_key: The label key, e.g. "Environment"
        :param label_value: The label key, e.g. "Production"
        :return: The id of the label with the provided label_key and label_value
        """
        endpoint = f'assets/labels/{label_key}/{label_value}'
        label_summary_object = self.centra_session.json_query(self.centra_session.urljoin_api(endpoint),
                                                              method='POST', data={"vms": asset_ids})
        return label_summary_object['id']

    def delete_label_by_name(self, label_name: str):
        """
        Delete a label by its name.
        :param label_name: The label name as a string, e.g. "App: Accounting" or "Environment: Production"
        """
        key, value = label_name.split(":")
        self.delete_label_by_key_value(key.strip(), value.strip())

    def delete_label_by_key_value(self, label_key, label_value):
        """
        Delete a label.
        :param label_key: The label key, e.g. "Environment"
        :param label_value: The label value, e.g. "Production"
        """
        label_ids = self.get_labels_ids(label_key, label_value)
        for label_id in label_ids:
            self.logger.debug(f"Trying to delete {label_id}")
            endpoint = f'visibility/labels/{label_id}'
            deleted_id = self.centra_session.json_query(
                self.centra_session.urljoin_api(endpoint),
                method='DELETE')
            assert deleted_id == label_id

    # TODO change signature to accept filt
    def get_labels_ids(self, label_key, label_value):
        """
        :param label_key: The label key, e.g. "Environment", "Role".
        :param label_value: The label's value, e.g. "Prod", "DB".
        :return: A list of label IDs (can be used for other API such as deleting labels by ID)
        """
        endpoint = f"visibility/labels"
        params = {
            "key": label_key,
            "value": label_value
        }
        # TODO move to generic get_labels with pagination
        list_of_matching_label_objects = self.centra_session.json_query(
            self.centra_session.urljoin_api(endpoint),
            method='GET',
            params=params
        )['objects']
        list_of_matching_label_ids = [x['id'] for x in list_of_matching_label_objects]
        self.logger.debug(f"Found {len(list_of_matching_label_ids)} matching labels")
        return list_of_matching_label_ids

    def get_system_notifications(self):
        return self.centra_session.json_query(
            self.centra_session.urljoin_api('system-notifications')
        )

    @property
    def is_connected(self) -> bool:
        """
        Use this to test for connectivity.
        :return: True if Centra is connected and answering the API.
        """
        try:
            return self.get_system_notifications() is not None
        except ManagementAPIError as e:
            self.logger.debug(f"Error which checking connectivity: {e}")
            return False



#!/usr/bin/env python

"""Tests for `centra_py_client` package."""
import pytest
import uuid

from unittest import TestCase
from unittest.mock import Mock, patch, call
from callee import Contains
from centra_py_client.exceptions import ManagementAPIError

from centra_py_client.centra_py_client import CentraClient
from centra_py_client.centra_session import CentraSession


class TestClient(TestCase):
    @patch("centra_py_client.centra_py_client.CentraSession.connect")
    @patch("centra_py_client.centra_py_client.CentraSession.json_query")
    def test_list_assets(self, mock_json_query, _):
        # arrange
        fake_asset = {"_id": "deadbeef-1337-1337-1337deadbeef1337"}
        fake_asset_2 = {"_id": "8eadbeef-1337-1337-1337deadbeef1338"}
        mock_json_query.return_value = {
            "objects": [fake_asset, fake_asset_2]
        }
        client = CentraClient(CentraSession("fakeaddr", "fakeuser", "fakepassword"))

        # act
        returned_assets = client.list_assets()

        # assert
        mock_json_query.assert_called_once()
        assert fake_asset in returned_assets

    @patch("centra_py_client.centra_py_client.CentraSession.connect")
    @patch("centra_py_client.centra_py_client.CentraSession.json_query")
    def test_add_label_to_assets(self, mock_json_query, _):
        # arrange
        fake_asset_id = str(uuid.uuid4())
        fake_asset_2_id = str(uuid.uuid4())
        fake_key = "key"
        fake_value = "value"

        mock_json_query.return_value = {
            "id": str(uuid.uuid4()),
            "key": fake_key,
            "value": fake_value,
            "name": f"{fake_key}: {fake_value}"
        }
        client = CentraClient(CentraSession("fakeaddr", "fakeuser", "fakepassword"))

        # act
        response = client.add_label_to_assets([fake_asset_id, fake_asset_2_id], fake_key, fake_value)

        # assert
        mock_json_query.assert_called_once_with(
            Contains(f'assets/labels/{fake_key}/{fake_value}'),
            method="POST",
            data={"vms": [fake_asset_id, fake_asset_2_id]})

    @patch("centra_py_client.centra_py_client.CentraSession.connect")
    @patch("centra_py_client.centra_py_client.CentraSession.json_query")
    def test_get_labels_ids(self, mock_json_query, _):
        id1 = uuid.uuid4()
        id2 = uuid.uuid4()
        mock_json_query.return_value = {
            'objects': [
                {"id": id1},
                {"id": id2},
            ]
        }

        client = CentraClient(CentraSession("fakeaddr", "fakeuser", "fakepassword"))
        key = "a_key"
        value = "a_value"
        label_ids = client.get_labels_ids(key, value)

        assert set(label_ids) == {id1, id2}
        mock_json_query.assert_called_once_with(
            Contains("visibility/labels"),
            method="GET",
            params={"key": key, "value": value}
        )

    @patch("centra_py_client.centra_py_client.CentraSession.connect")
    @patch("centra_py_client.centra_py_client.CentraClient.get_labels_ids")
    @patch("centra_py_client.centra_py_client.CentraSession.json_query")
    def test_delete_label_by_key_value(self, mock_json_query, mock_get_label_ids, _):
        first_id = "first_id"
        second_id = "second_id"
        mock_get_label_ids.return_value = [first_id, second_id]
        # This makes the call to json_query return the last part of the uri which is the label ID.
        # That's what the actual delete_label API returns.
        mock_json_query.side_effect = lambda uri, method: uri.split("/")[-1]

        client = CentraClient(CentraSession("fakeaddr", "fakeuser", "fakepassword"))
        key = "a_key"
        value = "a_value"
        client.delete_label_by_key_value(key, value)

        mock_json_query.assert_has_calls(
            calls=[
                call(Contains(first_id), method="DELETE"),
                call(Contains(second_id), method="DELETE")
            ],
            any_order=True
        )

    @patch("centra_py_client.centra_py_client.CentraSession.connect")
    @patch("centra_py_client.centra_py_client.CentraSession.json_query")
    def test_is_connected_sanity(self, mock_json_query, _):
        mock_json_query.return_value = {'total_count': 0, 'new_count': 0, 'items': []}
        client = CentraClient(CentraSession("fakeaddr", "fakeuser", "fakepassword"))
        assert client.is_connected
        mock_json_query.assert_called_once()

    @patch("centra_py_client.centra_py_client.CentraSession.connect")
    @patch("centra_py_client.centra_py_client.CentraSession.json_query")
    def test_is_connected_not_connected(self, mock_json_query, _):
        def raise_an_error(param1):
            raise ManagementAPIError(f"{param1} testerror")
        mock_json_query.side_effect = raise_an_error
        client = CentraClient(CentraSession("fakeaddr", "fakeuser", "fakepassword"))
        assert not client.is_connected
        mock_json_query.assert_called_once()



import datetime
import json
import logging
from typing import Dict, Union
from urllib.parse import urljoin

from centra_py_client.exceptions import ManagementAPIError, ManagementAPITimeoutError, RESTAuthenticationError

import requests
from requests.auth import AuthBase

MANAGEMENT_REST_API_PORT = 443
AUTHENTICATION_ERROR_HTTP_STATUS_CODE = 403
TIME_FORMAT_STRING = "%Y/%m/%d %H:%M:%S.%f"  # this should be parsable by dateutil.parser
REST_API_BASE_PATH_V3 = '/api/v3.0/'


class DatetimeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.strftime(TIME_FORMAT_STRING)
            # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


class JWTAuth(AuthBase):
    """Attaches JWT Authentication to the given Request object."""

    def __init__(self, token):
        # setup any auth-related data here
        if token is None:
            raise ManagementAPIError("REST Token not set!")
        self.token = token

    def __call__(self, r):
        # modify and return the request
        r.headers['Authorization'] = 'Bearer ' + self.token
        return r


class CentraSession:
    def __init__(
        self,
        management_address: str,
        auth_username: str,
        auth_password: str,
        base_api_path=REST_API_BASE_PATH_V3,
        verify_certificate: bool = True
    ):
        """
        Todo
        :param management_address:
        :param auth_username:
        :param auth_password:
        """
        self.logger = logging.getLogger(self.__class__.__name__)
        self.management_address = management_address
        self.auth_username = auth_username
        self.auth_password = auth_password

        self.http_server_root = f"https://{management_address}"
        self._requests_session = requests.Session()
        self._requests_session.verify = verify_certificate

        self.json_encoder = DatetimeEncoder()
        self.rest_auth_enabled = True
        self.authentication_handler = self.rest_authenticate

        self.set_base_api_path(base_api_path)

        self.connect()
        self.logger.debug(f"Connected to Centra successfully on {self.management_address}.")

    def set_base_api_path(self, base_api_path):
        # TODO import validators and validate the path
        self.base_api_path = base_api_path

    def urljoin_api(self, endpoint_path: str):
        return urljoin(self.base_api_path, endpoint_path)

    def rest_authenticate(self, rest_username, rest_password):
        """
        Perform JWT authentication through management REST API with username/password
        :param rest_username:
        :param rest_password:
        """
        self.logger.debug("REST Authenticating")
        response = self.json_query(uri=self.urljoin_api('authenticate'),
                                   method='POST',
                                   data={'username': rest_username, 'password': rest_password},
                                   authenticate=False)
        if 'access_token' in response:
            token = response['access_token']
            self.set_token(token)
        # todo raise exception when auth fails
        self.logger.debug("REST token obtained and set")

    def json_query(self,
                   uri,
                   method="GET",
                   data=None,
                   return_json=True,
                   params=None,
                   authenticate=True,
                   files=None,
                   convert_data_to_json=True) -> Union[bytes, Dict, str, None]:
        # TODO apijoin the uri
        if data is not None and convert_data_to_json:
            data = self.json_encoder.encode(data)
        response = self._query(uri=uri, method=method, data=data,
                               params=params, authenticate=authenticate, files=files)
        try:
            if not return_json:
                return response.content
            try:
                json_obj = json.loads(response.content)
            except TypeError:
                json_obj = json.loads(response.content.decode('utf-8'))
            if json_obj is not None and "code" in json_obj and 0 != json_obj["code"]:
                raise ManagementAPIError("Error: %s" % (json_obj["message"],))

            return json_obj
        except ValueError as exc:
            raise ManagementAPIError("Error reading server response: %s :: [%s]" % (str(exc), response.content))

    def _query(self, uri, method="GET", data=None, params=None, authenticate=True, files=None, **kwargs):
        if params is None:
            params = {}

        self.logger.debug("%s %s%s", method, uri, ('' if not params
                                                   else '?' + '&'.join("%s=%s" % (key, value)
                                                                       for key, value in params.items())))

        method_func = {"GET": self._requests_session.get,
                       "POST": self._requests_session.post,
                       "PUT": self._requests_session.put,
                       "PATCH": self._requests_session.patch,
                       "DELETE": self._requests_session.delete}[method]

        headers = {'content-type': 'application/json'} if files is None else None
        auth = JWTAuth(self.token) if (authenticate and self.rest_auth_enabled) else None
        # print urljoin(self.http_server_root, uri), data, headers, params
        try:
            r = method_func(urljoin(self.http_server_root, uri), data=data, headers=headers,
                            params=params, auth=auth, files=files, **kwargs)
        except requests.exceptions.RequestException as e:
            raise ManagementAPIError("Error while handling %s request for uri %s: %s" % (method, uri, e))

        if AUTHENTICATION_ERROR_HTTP_STATUS_CODE == r.status_code:  # This is a potential authorization error
            raise RESTAuthenticationError(r)

        if 200 != r.status_code:
            try:
                json_obj = json.loads(r.content)
            except:  # noqa: E722
                json_obj = r.content
                if isinstance(json_obj, bytes) and b"504 Gateway Time-out" in json_obj:
                    raise ManagementAPITimeoutError(json_obj)

            raise ManagementAPIError(json_obj)

        return r

    def connect(self):
        self.authentication_handler(self.auth_username, self.auth_password)

    def disconnect(self):
        self.json_query(self.urljoin_api('logout'), method='POST', return_json=False)

    def set_token(self, token):
        """
        Set JWT token, used for authentication with REST API
        :param token:
        """
        self.logger.debug("Setting REST token")
        self.token = token
        self._requests_session.auth = JWTAuth(self.token)  # so others can use this session



class ManagementAPIError(Exception):
    def __init__(self, message):
        super(ManagementAPIError, self).__init__(message)


class ManagementAPITimeoutError(ManagementAPIError):
    def __init__(self, message):
        super(ManagementAPITimeoutError, self).__init__(message)


class RESTAuthenticationError(ManagementAPIError):
    def __init__(self, response):
        data = response.json()
        super(RESTAuthenticationError, self).__init__('%s: %s' % (data['error'],
                                                                  data['description']))
        self.data = data



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = ['requests', ]  # todo read from requirements.txt

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]

setup(
    author="Yonatan Golan",
    author_email='yonatan.golan@guardicore.com',
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="Python client for Centra API access.",
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme,
    include_package_data=True,
    keywords='centra_py_client',
    name='centra_py_client',
    packages=find_packages(include=['centra_py_client', 'centra_py_client.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/guardicode/centra-py-client',
    version='0.3.0',
    zip_safe=False,
)
